#	OpenShot Video Editor is a program that creates, modifies, and edits video files.
#   Copyright (C) 2009  Jonathan Thomas
#
#	This file is part of OpenShot Video Editor (http://launchpad.net/openshot/).
#
#	OpenShot Video Editor is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	OpenShot Video Editor is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with OpenShot Video Editor.  If not, see <http://www.gnu.org/licenses/>.

import gtk
import goocanvas
import os
import uuid

########################################################################
class clip:
	"""This class represents a media clip on the timeline."""

	#----------------------------------------------------------------------
	def __init__(self, clip_name, color, position_on_track, start_time, end_time, parent_track, file_object):
		"""Constructor"""
		
		# init variables for clip object
		self.name = clip_name		   # the name of the clip
		self.color = color			  # the color of the clip, used to organize clips
		self.start_time = start_time	# the time in seconds where we start playing a clip
		self.end_time = end_time		# the time in seconds where we stop playing a clip
		self.max_length = self.length()		# this is the max length of the clip in seconds
		self.position_on_track = float(position_on_track)  # the time in seconds to start playing the clip relative to the track
		self.playback_rate = 1.0		# the rate of playback (this will change if you want to slow down or speed up the clip)
		self.parent = parent_track	  # the parent track this clip lives on
		self.file_object = file_object  # the file object that this clip is linked to
		self.play_video = True
		self.play_audio = True
		self.unique_id = str(uuid.uuid1())	
		
		# For example:  imagine a clip that is 30 seconds long.  If we wanted to only play a 10 second section (from 15 to 25 
		# second range) of this clip, and we wanted the 10 second section to start playing 3 seconds into the tracks timeline, 
		# here are the settings you would need on the clip:
		#
		# REMEMBER: All values are in seconds (of type float... and thus they contain decimals)
		# length = 10.0
		# start_time = 15.0
		# end_time = 25.0
		# position_on_track = 3.0
		
		
	def length(self):
		# calculate the length of this clip (in decimal seconds)
		length = self.end_time - self.start_time
		
		# return length
		return length
		

	
	def Render(self):

		# Render the clip to the timeline
		self.RenderClip()
	 
	 
	def GenerateXML(self, dom, xmlParentNode):
		
		# get the frames per second (from the project)
		fps = self.parent.parent.project.fps()
		
		# get the root tractor node 
		tractor_node = dom.getElementsByTagName("tractor")[0]
		
		# determine how much blank space to insert (if any)
		blank = self.get_blank_space() * fps
		if blank:
			blank_node = dom.createElement("blank")
			blank_node.setAttribute("length", str(blank))
			xmlParentNode.appendChild(blank_node)
		
		# calculate IN and OUT frames (based on the # of seconds)
		in_frame_number = self.start_time * fps
		out_frame_number = self.end_time * fps
		
		# create the clip producer node
		producer = dom.createElement("producer")
		producer.setAttribute("id", self.name)
		producer.setAttribute("in", str(in_frame_number))
		producer.setAttribute("out", str(out_frame_number))
		#producer.setAttribute("max_frames", str(self.file_object.max_frames))
		#producer.setAttribute("project_fps", str(fps))
		#producer.setAttribute("fps", str(self.file_object.fps))
		#producer.setAttribute("start_time", str(self.start_time))
		#producer.setAttribute("end_time", str(self.end_time))
		xmlParentNode.appendChild(producer)
		
		# add the resource to the producer node
		property = dom.createElement("property")
		property.setAttribute("name", "resource")
		text = dom.createTextNode(self.file_object.location)
		property.appendChild(text)
		producer.appendChild(property)
		
		# Get the track index (i.e. track 0, 1, 2, 3, etc...)
		track_index = self.parent.parent.tracks.index(self.parent)
		
		# Flip the index (since we want 0 on the bottom)
		track_index_flipped = (len(self.parent.parent.tracks) - track_index) - 1
		
		# Add VOLUME filter (for audio mute)
		if self.play_audio == False or self.parent.play_audio == False:
			# create Volume Filter
			# <filter mlt_service="volume" track="0" in="0" out="30" gain="0"/>
			
			# get the IN / OUT frame for the Transistion (absolute frame # of the project)
			trans_in_frame_number = self.position_on_track * fps
			trans_out_frame_number = trans_in_frame_number + (self.end_time * fps)
			
			filter = dom.createElement("filter")
			filter.setAttribute("mlt_service", "volume")
			filter.setAttribute("in", str(trans_in_frame_number))
			filter.setAttribute("out", str(trans_out_frame_number))
			filter.setAttribute("track", str(track_index_flipped))
			filter.setAttribute("gain", "0")
			
			# add volume filter to tractor
			tractor_node.appendChild(filter)
			
		
		# Create the COMPOSITE Transition for this clip.  This allows transparency
		# to work between the different tracks.  Only needed on tracks that are not zero.
		if track_index_flipped > 0:
			##############################
			#         COMPOSITE
			##############################
			# get the IN / OUT frame for the Transistion (absolute frame # of the project)
			trans_in_frame_number = self.position_on_track * fps
			trans_out_frame_number = trans_in_frame_number + (self.end_time * fps)
			
			transition = dom.createElement("transition")
			transition.setAttribute("in", str(trans_in_frame_number))
			transition.setAttribute("out", str(trans_out_frame_number))
			
			property = dom.createElement("property")
			property.setAttribute("name", "mlt_service")
			text = dom.createTextNode("composite")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "a_track")
			text = dom.createTextNode("0")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "b_track")
			text = dom.createTextNode(str(track_index_flipped))
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "geometry")
			# hide the video (if needed)
			if self.play_video == False or self.parent.play_video == False:
				text = dom.createTextNode("0=0%,0%:100%x100%:0; -1=0%,0%:100%x100%:0")
			else:
				text = dom.createTextNode("0=0%,0%:100%x100%:100; -1=0%,0%:100%x100%:100")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "halign")
			text = dom.createTextNode("centre")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "valign")
			text = dom.createTextNode("centre")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "distort")
			text = dom.createTextNode("1")
			property.appendChild(text)
			transition.appendChild(property)
			
			# add composite transition to tractor
			tractor_node.appendChild(transition)
			
			##############################
			#			 MIX
			##############################
			
			transition = dom.createElement("transition")
			transition.setAttribute("in", str(trans_in_frame_number))
			transition.setAttribute("out", str(trans_out_frame_number))
			
			property = dom.createElement("property")
			property.setAttribute("name", "mlt_service")
			text = dom.createTextNode("mix")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "a_track")
			text = dom.createTextNode("0")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "b_track")
			text = dom.createTextNode(str(track_index_flipped))
			property.appendChild(text)
			transition.appendChild(property)
			
#			property = dom.createElement("property")
#			property.setAttribute("name", "geometry")
#			text = dom.createTextNode("0=0%,0%:100%x100%:100; -1=0%,0%:100%x100%:100")
#			property.appendChild(text)
#			transition.appendChild(property)
			
			# These 2 properties combine the audio from 2 tracks equally
			property = dom.createElement("property")
			property.setAttribute("name", "combine")
			text = dom.createTextNode("1")
			property.appendChild(text)
			transition.appendChild(property)
			
			property = dom.createElement("property")
			property.setAttribute("name", "always_active")
			text = dom.createTextNode("1")
			property.appendChild(text)
			transition.appendChild(property)
			
			# These 2 properties allow cross-fading the 2 tracks
#			property = dom.createElement("property")
#			property.setAttribute("name", "start")
#			text = dom.createTextNode("0.5")
#			property.appendChild(text)
#			transition.appendChild(property)
#			
#			property = dom.createElement("property")
#			property.setAttribute("name", "end")
#			text = dom.createTextNode("0.5")
#			property.appendChild(text)
#			transition.appendChild(property)

			# add composite transition to tractor
			tractor_node.appendChild(transition)
			



	#----------------------------------------------------------------------
	def RenderClip(self, item = None, x_offset = 0):
		"""This adds a clip to the canvas with 3 images: a left, middle, and right.  If an goocanvas item is passed in,
		this method will simply re-size the inner images.  If no goocanvas item is passed in, it will create one. The reason
		this method is slightly complicated is due to the resize / trim clip feature, since it has to toggle between the small
		version of a clip and the 3 image version."""

		# get the previous track from the parent sequence (if any)
		pixels_per_second = self.parent.parent.get_pixels_per_second()
		x = float(self.position_on_track * pixels_per_second) + x_offset
		y = self.parent.y_top + 2

		# get a reference to the 2 main canvas objects & theme
		theme = self.parent.parent.project.theme

		# load clip images
		imgTrack_Middle = gtk.image_new_from_file("%s/themes/%s/Clip_Middle_%s.png" % (self.parent.parent.project.form.openshot_path, theme, self.color))
		imgTrack_Left = gtk.image_new_from_file("%s/themes/%s/Clip_Left_%s.png" % (self.parent.parent.project.form.openshot_path, theme, self.color))
		imgTrack_Right = gtk.image_new_from_file("%s/themes/%s/Clip_Right_%s.png" % (self.parent.parent.project.form.openshot_path, theme, self.color))
		
		# get height & width of left image
		imgTrack_Left_Height = imgTrack_Left.get_pixbuf().get_height()
		imgTrack_Left_Width = imgTrack_Left.get_pixbuf().get_width()
		imgTrack_Right_Width = imgTrack_Right.get_pixbuf().get_width()
		
		# Get Size of Window (to determine how wide the middle image should be streched)
		Size_Of_Middle = (self.length() * pixels_per_second) - (imgTrack_Left_Width + imgTrack_Right_Width) + 6
		total_pixel_length = int(self.length() * pixels_per_second) + 1
		has_small_images = False
		has_regular_images = False
		
		# Get the canvas group for this clip
		if item:
			# set the group object
			GroupClip = item
			
			# set the unique ID of the group
			GroupClip.set_data ("id", self.unique_id)
			
			if self.get_canvas_child(item, "small_middle"):
				has_small_images = True
				
			if self.get_canvas_child(item, "left"):
				has_regular_images = True
				
		else:
			# canvas group is not passed in
			# Get root group of the canvas
			canvas_right = self.parent.parent.project.form.MyCanvas
			root_right = canvas_right.get_root_item ()		
			
			# Create the Group (for the clip)
			GroupClip = goocanvas.Group (parent = root_right)
			
			# set the unique ID of the group
			GroupClip.set_data ("id", self.unique_id)
			
			# connect drag n drop events to the new cavnas group
			GroupClip.connect ("motion_notify_event", self.on_motion_notify_x)
			GroupClip.connect ("button_press_event", self.on_button_press_x)
			GroupClip.connect ("button_release_event", self.on_button_release_x)
		
		

		# for clips that are too small to use the 3 image approach, just use
		# the middle image, and don't add a thumbnail
		if total_pixel_length < imgTrack_Left_Width + imgTrack_Right_Width + 60:

			# ///////////////////////////////////////////////////////
			# Add ONLY the MIDDLE Image to Group
			# ///////////////////////////////////////////////////////

			# Resize Middle pixbuf to be the entire length of the clip
			middlePixBuf = imgTrack_Middle.get_pixbuf()
			pixbuf_list = self.parent.parent.split_images(middlePixBuf, imgTrack_Left_Height, total_pixel_length)

			# Add Middle Image to Group (this can be multiple image tiled together
			pixbuf_x = 0
			for pixbuf in pixbuf_list:
				# get width of this pixbuf
				pixbuf_width = pixbuf.get_width()
				pixbuf_height = pixbuf.get_height()
				
				# Add the middle image
				if item == None or has_small_images == False:
					# Create the small image items (since they don't exist yet)
					# create middle canvas image object
					image2 = goocanvas.Image (parent = GroupClip,
										  pixbuf = pixbuf,
										  x = x + pixbuf_x,
										  y = y)  
					image2.set_data ("id", "small_middle")
					
					# Add a translucent blue rectab
					rec2 = goocanvas.Rect (parent = GroupClip,
										  x = x + pixbuf_x,
										  y = y,
										  width = pixbuf_width,
										  height = pixbuf_height,
										  line_width = 1,
										  stroke_color_rgba = 1401402500)
					rec2.set_data ("id", "small_rect")
					
					# create new canvas text object
					if has_regular_images == False:
						
						text_width = pixbuf_width
						if text_width - 8 <= 15:
							text_width = 1
						else:
							text_width = text_width - 8
						
						text1 = goocanvas.Text (parent = GroupClip,
											text = "%s" % self.name,
											font = "Sans 8",
											antialias = False,
											x = x + 8,
											y = y + 2,
											width = text_width,
											ellipsize = 3)
						text1.set_data ("id", "clip_name")
						
						# determine if text has enough width to be visible
						if text_width == 1:
							text1.set_properties(visibility = 1)
						
					else:
						# get existing text
						text1 = self.get_canvas_child(item, "clip_name")

						
				elif has_small_images:
					# resize existing canvas image object
					self.get_canvas_child(item, "small_middle").set_properties(visibility = 2)
					self.get_canvas_child(item, "small_middle").set_properties(pixbuf = pixbuf)
					self.get_canvas_child(item, "small_middle").set_properties(x = x + pixbuf_x)
					
					self.get_canvas_child(item, "small_rect").set_properties(visibility = 2)
					self.get_canvas_child(item, "small_rect").set_properties(x = x + pixbuf_x)
					self.get_canvas_child(item, "small_rect").set_properties(width = pixbuf_width)
					
					text_width = pixbuf_width
					if text_width - 8 <= 15:
						text_width = 1
						# hide clip name
						self.get_canvas_child(item, "clip_name").set_properties(visibility = 1)
					else:
						text_width = text_width - 8
						# show clip name
						self.get_canvas_child(item, "clip_name").set_properties(visibility = 2)
						
					# resize clip name
					self.get_canvas_child(item, "clip_name").set_properties(x = x + 8)
					self.get_canvas_child(item, "clip_name").set_properties(width = text_width)
				
				
				# hide the left, right, and middle, text, and thumbnail items (if any)
				if has_regular_images:
					self.get_canvas_child(item, "middle").set_properties(visibility = 1)
					self.get_canvas_child(item, "left").set_properties(visibility = 1)
					self.get_canvas_child(item, "right").set_properties(visibility = 1)
					self.get_canvas_child(item, "thumbnail").set_properties(visibility = 1)
					self.get_canvas_child(item, "clip_visible").set_properties(visibility = 1)
					self.get_canvas_child(item, "clip_speaker").set_properties(visibility = 1)
					#self.get_canvas_child(item, "clip_name").set_properties(visibility = 1)
					
					# raise the clip text (since it can get hidden by other clip-related layers)
					self.get_canvas_child(item, "clip_name").raise_(None)
					

				# increment the x
				pixbuf_x = pixbuf_x + pixbuf_width
			
		else:
			
			# hide the small version of the clip (if any)
			if has_small_images:
				
				# hide the smaller version, if it exists
				self.get_canvas_child(item, "small_middle").set_properties(visibility = 1)
				self.get_canvas_child(item, "small_rect").set_properties(visibility = 1)


			# ///////////////////////////////////////////////////////
			# Add MIDDLE Image to Group
			# ///////////////////////////////////////////////////////

			# Resize Middle pixbuf
			middlePixBuf = imgTrack_Middle.get_pixbuf()
			pixbuf_list = self.parent.parent.split_images(middlePixBuf, imgTrack_Left_Height, int(Size_Of_Middle))			 
			
			# Add Middle Image to Group (this can be multiple image tiled together
			pixbuf_x = 0
			for pixbuf in pixbuf_list:
				# get width of this pixbuf
				pixbuf_width = pixbuf.get_width()
				
				if has_regular_images == False:
					# create middle canvas image object
					image2 = goocanvas.Image (parent = GroupClip,
										  pixbuf = pixbuf,
										  x = (x + imgTrack_Left_Width - 3) + pixbuf_x,
										  y = y) 
					image2.set_data ("id", "middle")
					
				else:
					# get existing object
					existing_item = self.get_canvas_child(item, "middle")
					
					# resize existing canvas image object
					existing_item.set_properties(pixbuf = pixbuf)
					existing_item.set_properties(x = (x + imgTrack_Left_Width - 3) + pixbuf_x)
					existing_item.set_properties(visibility = 2)
				
				# increment the x
				pixbuf_x = pixbuf_x + pixbuf_width									  

				
			# ///////////////////////////////////////////////////////
			# Add LEFT Image to Group
			# ///////////////////////////////////////////////////////
			if has_regular_images == False:

				# create canvas image object
				imageLeft = goocanvas.Image (parent = GroupClip,
									  pixbuf = imgTrack_Left.get_pixbuf(),
									  x = x,
									  y = y)  

				imageLeft.set_data ("id", "left")
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "left")
				
				# resize existing canvas image object
				existing_item.set_properties(x = x)
				existing_item.set_properties(visibility = 2)
				
				

			# ///////////////////////////////////////////////////////
			# Add RIGHT Image to Group
			# ///////////////////////////////////////////////////////
			if has_regular_images == False:
				
				# create canvas image object
				imageRight = goocanvas.Image (parent = GroupClip,
									  pixbuf = imgTrack_Right.get_pixbuf(),
									  x = x + imgTrack_Left_Width + Size_Of_Middle - 5,
									  y = y)
				
				imageRight.set_data ("id", "right")
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "right")
				
				# resize existing canvas image object
				existing_item.set_properties(x = x + imgTrack_Left_Width + Size_Of_Middle - 5)
				existing_item.set_properties(visibility = 2)
				

				
			# ///////////////////////////////////////////////////////
			# Add THUMBNAIL Image to Group
			# ///////////////////////////////////////////////////////	 
			if has_regular_images == False:
				# get the thumbnail image
				pbThumb = self.file_object.get_thumbnail(38, 35)
				
				# create canvas image object
				imgThumb = goocanvas.Image (parent = GroupClip,
									pixbuf = pbThumb,
									x = x + 6,
									y = y + 16)
				
				imgThumb.set_data ("id", "thumbnail")
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "thumbnail")
				
				# re-position existing canvas image object
				existing_item.set_properties(x = x + 6)
				existing_item.set_properties(visibility = 2)
				
				
			# ///////////////////////////////////////////////////////
			# Add BUTTONS Image to Group
			# ///////////////////////////////////////////////////////	
			
			# *** VISIBLE BUTTONS
			if has_regular_images == False:
			
				# Load buttons
				if self.play_video:
					imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/visible_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
				else:
					imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/not_visible_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
					
				# Add Visible Image to Group
				image5 = goocanvas.Image (parent = GroupClip,
										  pixbuf = imgTrack_Visible.get_pixbuf(),
										  x = x + 46,
										  y = y + 18)  
				
				# Track the state of the hover over image
				image5.set_data ("id", "clip_visible")
				image5.connect ("button_press_event", self.on_visible_click)
				
				# Get the track index (i.e. track 0, 1, 2, 3, etc...)
				track_index = self.parent.parent.tracks.index(self.parent)
				
				# Flip the index (since we want 0 on the bottom)
				track_index_flipped = (len(self.parent.parent.tracks) - track_index) - 1
				
				# Hide if track 0
				if track_index_flipped == 0:
					image5.set_properties(visibility = 1)
					
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "clip_visible")
				
				# resize existing canvas image object
				existing_item.set_properties(x = x + 46)
				existing_item.set_properties(visibility = 2)
				
				# Get the track index (i.e. track 0, 1, 2, 3, etc...)
				track_index = self.parent.parent.tracks.index(self.parent)
				
				# Flip the index (since we want 0 on the bottom)
				track_index_flipped = (len(self.parent.parent.tracks) - track_index) - 1
				
				# Hide if track 0
				if track_index_flipped == 0:
					existing_item.set_properties(visibility = 1)
				
				
			# *** AUDIO BUTTONS
			if has_regular_images == False:
			
				# Load buttons
				if self.play_audio:
					imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/speaker_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
				else:
					imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/speaker_mute_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
					
				# Add Visible Image to Group
				image5 = goocanvas.Image (parent = GroupClip,
										  pixbuf = imgTrack_Visible.get_pixbuf(),
										  x = x + 66,
										  y = y + 18)  
				
				# Track the state of the hover over image
				image5.set_data ("id", "clip_speaker")
				image5.connect ("button_press_event", self.on_audio_click)
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "clip_speaker")
				
				# resize existing canvas image object
				existing_item.set_properties(x = x + 66)
				existing_item.set_properties(visibility = 2)

	
				
			# ///////////////////////////////////////////////////////
			# Add TEXT Image to Group
			# ///////////////////////////////////////////////////////
			
			if item == None:
				# create new canvas text object
				text1 = goocanvas.Text (parent = GroupClip,
									text = "%s" % self.name,
									font = "Sans 8",
									antialias = False,
									x = x + 8,
									y = y + 2,
									width = Size_Of_Middle,
									ellipsize = 3)
				
				text1.set_data ("id", "clip_name")
				
			else:
				# get existing object
				existing_item = self.get_canvas_child(item, "clip_name")				
				
				# re-position existing canvas image object
				existing_item.set_properties(x = x + 8)
				existing_item.set_properties(visibility = 2)
				existing_item.set_properties(text = "%s (%s - %s)" % (self.name, self.start_time, self.end_time))
				existing_item.set_properties(width = Size_Of_Middle)
				
				# raise the clip text (since it can get hidden by other clip-related layers)
				existing_item.raise_(None)

				

		
		# return the clip... this is useful when a new clip is dropped onto the canvas
		return GroupClip
	
	
	
	def get_canvas_child(self, group, requested_child_id):
		"""this method loops though the children objects of this group looking 
		for the item with a specfic id."""
		
		for index in range(0, group.get_n_children()):
			child = group.get_child(index)
			child_id = child.get_data ("id")

			if child_id == requested_child_id:
				return child
			
		return None
	
	
	def protect_visible_button(self):
		# Get root group of the canvas
		canvas_right = self.parent.parent.project.form.MyCanvas
		root_right = canvas_right.get_root_item ()
		
		# Get the track index (i.e. track 0, 1, 2, 3, etc...)
		track_index = self.parent.parent.tracks.index(self.parent)
		
		# Flip the index (since we want 0 on the bottom)
		track_index_flipped = (len(self.parent.parent.tracks) - track_index) - 1

		# Hide if track 0
		if track_index_flipped == 0:
			
			# loop through each clip on this track
			for clip in self.parent.clips:

				# get the Clip Group
				ClipGroup = self.get_canvas_child(root_right, clip.unique_id)

				# HIDE VISIBLE BUTTON
				existing_item = self.get_canvas_child(ClipGroup, "clip_visible")
				if existing_item:
					existing_item.set_properties(visibility = 1)
				
		else:
			
			# loop through each clip on this track
			for clip in self.parent.clips:

				# get the Clip Group
				ClipGroup = self.get_canvas_child(root_right, clip.unique_id)

				# SHOW VISIBLE BUTTON
				existing_item = self.get_canvas_child(ClipGroup, "clip_visible")
				if existing_item:
					existing_item.set_properties(visibility = 2)
	


	def on_motion_notify_x (self, item, target, event):
		"""this method allows the clip to be dragged and dropped on a track"""	  

		# get the new x,y coordinates from the mouse
		new_x = event.x
		new_y = event.y	
		
		# get the pixels per second from the parent sequence
		pixels_per_second = self.parent.parent.get_pixels_per_second()
		
		# determine end pixel of sequence
		end_of_timeline = self.parent.parent.length * pixels_per_second
		
		
		# determine what cursor mode is enable (arrow, razor, snap, etc...)
		(isArrow, isRazor, isSnap, isResize) = self.parent.parent.project.form.get_toolbar_options()
		
		# ARROW MODE
		if isArrow:
			
			# Move the clip based on the x, y of the mouse
			if (event.state & gtk.gdk.BUTTON1_MASK):

				# don't allow the clip to slide past the beginning of the canvas
				total_x_diff = new_x - self.drag_x 
				total_y_diff = event.y - self.drag_y
				if (item.get_bounds().x1 + total_x_diff < 0):
					total_x_diff = 0 - item.get_bounds().x1
				elif (item.get_bounds().x2 + total_x_diff > end_of_timeline):
					total_x_diff = end_of_timeline - item.get_bounds().x2
	
				# be sure that the clip is being dragged over a valid drop target (i.e. a track)
				if self.is_valid_drop(item.get_bounds().x1 + total_x_diff, item.get_bounds().y1 + total_y_diff):
	
					# move clip				
					item.translate (total_x_diff, total_y_diff)
					
		# RESIZE MODE
		if isResize:
			
			
			# update cursor
			self.parent.parent.project.form.current_cursor[1] = int(event.x_root)
			self.parent.parent.project.form.current_cursor[2] = int(event.y_root)
			self.parent.parent.project.form.current_cursor[3] = "clip"
			
			if (event.state & gtk.gdk.BUTTON1_MASK):

				# get the direction from the cursor object
				direction = self.parent.parent.project.form.current_cursor[0]
				
			else:
				# only calculate LEFT or RIGHT when the mouse is not clicked.  Once the user starts resizing
				# a clip, we want to just use the direction in the cursor object.  In other words, we can't allow
				# the direction to change while we are resizing.
				# determine if user is resizing the LEFT or RIGHT side of the clip
				center_of_clip = (self.position_on_track * pixels_per_second) + ((self.length() * pixels_per_second) / 2)
				direction = "left"
				if event.x_root < center_of_clip:
					direction = "left"
				else:
					direction = "right"
				
				# if right, update the cursor the "right grab" icon
				if direction == "right":
					# update cursor variable
					if self.parent.parent.project.form.current_cursor[0] != "right":
	
						# change cursor to "right bar"
						self.parent.parent.project.form.current_cursor[0] = "right"
						self.parent.parent.project.form.MyCanvas.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.RIGHT_SIDE)) 
	
				# if left, update the cursor the "left grab" icon
				if direction == "left":
					# update cursor variable
					if self.parent.parent.project.form.current_cursor[0] != "left":
						
						# change cursor to "left bar"
						self.parent.parent.project.form.current_cursor[0] = "left" 
						self.parent.parent.project.form.MyCanvas.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.LEFT_SIDE))

				
				
			
			# Move the clip based on the x, y of the mouse (if the mouse button is pressed)
			if (event.state & gtk.gdk.BUTTON1_MASK):
				
				# Get side of clip that is being re-sized
				side = self.parent.parent.project.form.current_cursor[0]  # left or right
				
				# Get the x and y difference
				original_end_pos = (self.original_start_pos * pixels_per_second) + (self.original_length * pixels_per_second)
				x_offset = event.x - event.x_root

				if side == "left":
					# calculate new start position and length
					new_position = float(event.x_root) / float(pixels_per_second)
					new_length = float(original_end_pos - event.x_root) / pixels_per_second
					new_start_time = self.end_time - new_length

					# stop the clip from going too far to the left
					if new_start_time < 0:
						new_position = new_position - new_start_time
						new_length = self.end_time
						new_start_time = self.end_time - new_length
						
						
					# update the properties of this clip
					if new_position >= 0 and new_length >= 0 and new_length <= self.max_length and new_start_time >= 0:
						
						self.position_on_track = new_position
						#self.length() = new_length
						self.start_time = new_start_time
						
						# adjust x for windows
						if os.name == 'nt':
							x_offset = x_offset - 2
							
						# re-render this clip to the canvas (passing the canvas group to the RenderClip method)
						self.RenderClip(item, x_offset)

				else:
					# RIGHT side
					# calculate the new clip length
					new_length = (event.x_root - (self.position_on_track * pixels_per_second)) / pixels_per_second
					new_end_time = self.start_time + new_length

					# prevent the clip from resizing too far to the right
					if self.start_time + new_length > self.max_length:
						new_end_time = self.max_length
						new_length = new_end_time - self.start_time

					# update the properties of this clip
					if new_length > 0 and new_end_time > 0 and new_length <= self.max_length and new_end_time <= self.max_length:
						#self.length() = new_length
						self.end_time = new_end_time

						# re-render this clip to the canvas (passing the canvas group to the RenderClip method)
						self.RenderClip(item, x_offset)

				
		return True
	

	def update(self, x, y, parent_track):
			""" This method updates the x and y settings of the clip in the object model """
			# get the pixels per second from the parent sequence
			pixels_per_second = self.parent.parent.get_pixels_per_second()

			# update the position of the clip   
			self.position_on_track = float(x) / float(pixels_per_second)

			# remove clip from current parent track (if new parent is different)
			if (parent_track != self.parent):
				# remove clip				
				self.parent.clips.remove(self)
				
				# add to new parent track
				parent_track.clips.append(self)
				self.parent = parent_track

	
	
	def on_button_press_x (self, item, target, event):
		""" This method initializes some variables needed for dragging and dropping a clip """
		# raise the group up to the top level
		item.raise_(None)
		self.parent.parent.play_head.raise_(None)		
		
		# set the x and y where the cursor started dragging from
		self.drag_x = event.x
		self.drag_y = event.y
		
		# raise playhead above clips
		self.parent.parent.play_head.raise_(None)

		# only respond to the first mouse button
		if event.button == 1:
			# determine what cursor mode is enable (arrow, razor, snap, etc...)
			(isArrow, isRazor, isSnap, isResize) = self.parent.parent.project.form.get_toolbar_options()
			
			# ARROW MODE
			if isArrow:
				
				if event.state & gtk.gdk.SHIFT_MASK:
					# remove clip from goocanvas
					parent = item.get_parent()
					child_num = parent.find_child (item)
					parent.remove_child (child_num)
					
					# remove clip from parent track
					self.parent.clips.remove(self)
				else:
					# change the cursor for the drag n drop operation
					fleur = gtk.gdk.Cursor (gtk.gdk.FLEUR)
					canvas = item.get_canvas ()
					canvas.pointer_grab (item, gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.BUTTON_RELEASE_MASK, fleur, event.time)

			
			# RAZOR MODE
			elif isRazor:
				# Divide the clip into 2 clips at the x position of the cursor
				self.divide_clip(event.x_root, item)
				pass

			# RESIZE MODE
			elif isResize:
				# remember the original length and position before resizing starts
				self.original_length = self.length()
				self.original_start_pos = self.position_on_track
				self.original_clip_in = self.start_time
				self.original_clip_out = self.end_time
				pass

			# SNAP MODE
			elif isSnap:
				pass
			
		elif event.button == 3:
			# show the track popup menu
			self.parent.parent.project.form.mnuClip1.showmnu(event, self)
 
		return True
	
	
	def divide_clip(self, x, canvas_item):
		"""Divide a clip into to smaller clips at a specific x coordinate"""
		# get the pixels per second from the parent sequence
		pixels_per_second = self.parent.parent.get_pixels_per_second()
		seconds_for_x = x / pixels_per_second
		original_clip_length = self.length()
		original_end_time = self.end_time

		# remove clip from goocanvas
		parent = canvas_item.get_parent()
		child_num = parent.find_child (canvas_item)
		parent.remove_child (child_num)
		
		#print " ------ START CUT -------- "
		#print "x: %s" % x
		#print "original_clip_length: %s" % original_clip_length
		#print "original_end_time: %s" % original_end_time
		#print "seconds_for_x: %s" % seconds_for_x
		#print "self.start_time: %s" % self.start_time
		#print "self.max_length: %s" % self.max_length
		
		
		# Modify the 1st clip, and render it to the screen
		self.end_time = (seconds_for_x - self.position_on_track) + self.start_time
		self.RenderClip()
		
		
		#print "self.end_time: %s" % self.end_time
		#print "self.length(): %s" % self.length()
		#print "\n ------- MAKE 2ND CLIP ------- "
		
		# calculate the 2nd clip properties
		new_start_time = self.end_time
		new_end_time = original_end_time
		
		#print "new_start_time: %s" % new_start_time
		#print "new_end_time: %s" % new_end_time

		# Create the 2nd clip object
		SecondClip = self.parent.AddClip(self.name, self.color, seconds_for_x, new_start_time, new_end_time, self.file_object)
		SecondClip.max_length = self.max_length
		SecondClip.RenderClip()
		
		# re-order clip objects on track (since they might be out of order now)
		self.parent.reorder_clips()
					

	def get_snap_difference(self, clip_object, canvas_item):
		"""Determine the number of pixels to shift this clip to snap to it's 
		closest neighbor clip (if any)"""

		# get the pixels per second from the parent sequence
		pixels_per_second = clip_object.parent.parent.get_pixels_per_second()
		old_x = clip_object.position_on_track * pixels_per_second  # get the old x coordinate of the clip (used to determine direction)

		# get the index of this clip (from the parent collection)
		clip_index = clip_object.parent.clips.index(clip_object)
		clip_length = clip_object.length() * pixels_per_second
		closest_clip = None
		distance_from_clip = 0.0
		distance_from_left_clip = 0.0
		distance_from_right_clip = 0.0
		
		# determine the direction of the drag
		direction = ""
		if canvas_item.get_bounds().x1 < old_x:
			direction = "left"
		else:
			direction = "right"
		
		# if there is a clip to the left of this one, find the distance to it's edge
		if (clip_index - 1) >= 0 and (clip_index - 1) < len(clip_object.parent.clips):
			closest_clip = clip_object.parent.clips[clip_index - 1]
			clip_x = closest_clip.position_on_track * pixels_per_second
			closest_clip_length = closest_clip.length() * pixels_per_second
			closest_clip_position = clip_x + closest_clip_length
			distance_from_left_clip = closest_clip_position - canvas_item.get_bounds().x1

		# if there is a clip to the right of this one, find the distance to it's edge
		if (clip_index + 1) >= 0 and (clip_index + 1) < len(clip_object.parent.clips):
			closest_clip = clip_object.parent.clips[clip_index + 1]
			closest_clip_position = closest_clip.position_on_track * pixels_per_second
			distance_from_right_clip = closest_clip_position - (canvas_item.get_bounds().x1 + clip_length)

		# limit the left /right snapping to 10 pixels
		if distance_from_left_clip > 10 or distance_from_left_clip < -10:
			distance_from_left_clip = 0.0
		if distance_from_right_clip > 10 or distance_from_right_clip < -10:
			distance_from_right_clip = 0.0

		# determine which direction and what clip to snap to (based on the direction the clip is moving)
		if direction == "left" and distance_from_left_clip != 0:
			distance_from_clip = distance_from_left_clip
		elif direction == "right" and distance_from_right_clip != 0:
			distance_from_clip = distance_from_right_clip
		elif distance_from_left_clip != 0:
			distance_from_clip = distance_from_left_clip
		elif distance_from_right_clip != 0:
			distance_from_clip = distance_from_right_clip

		# return the # of pixesl to snap the clip
		return distance_from_clip
	
	
	def get_blank_space(self):
		"""Determine the number of pixels to shift this clip to snap to it's 
		closest neighbor clip (if any)"""

		# get the pixels per second from the parent sequence
		pixels_per_second = self.parent.parent.get_pixels_per_second()
		old_x = self.position_on_track * pixels_per_second  # get the old x coordinate of the clip (used to determine direction)

		# get the index of this clip (from the parent collection)
		clip_index = self.parent.clips.index(self)
		clip_length = self.length() * pixels_per_second
		closest_clip = None
		distance_from_clip = 0.0
		distance_from_left_clip = 0.0
		
		# determine the direction of the drag
		direction = "left"
		
		# if there is a clip to the left of this one, find the distance to it's edge
		if (clip_index - 1) >= 0 and (clip_index - 1) < len(self.parent.clips):
			closest_clip = self.parent.clips[clip_index - 1]
			clip_x = closest_clip.position_on_track * pixels_per_second
			closest_clip_length = closest_clip.length() * pixels_per_second
			closest_clip_position = clip_x + closest_clip_length
			distance_from_left_clip = float((old_x - closest_clip_position) / pixels_per_second)
		else:
			distance_from_left_clip = float((old_x) / pixels_per_second)
		
		# zero out any very, very small differences
		if distance_from_left_clip < .01:
			distance_from_left_clip = 0

		# determine which direction and what clip to snap to (based on the direction the clip is moving)
		if direction == "left" and distance_from_left_clip != 0:
			distance_from_clip = distance_from_left_clip
		elif distance_from_left_clip != 0:
			distance_from_clip = distance_from_left_clip

		# return the # of pixesl to snap the clip
		return distance_from_clip
	

	def on_button_release_x (self, item, target, event):
		""" This method drops a clip, and snaps the clip to the nearest valid track """

		# raise the play head to the top
		self.parent.parent.play_head.raise_(None)

		# get reference to the canvas, and stop dragging the item
		canvas = item.get_canvas()
		canvas.pointer_ungrab (item, event.time)
		
		# determine what cursor mode is enable (arrow, razor, snap, etc...)
		(isArrow, isRazor, isSnap, isResize) = self.parent.parent.project.form.get_toolbar_options()
		
		if isArrow:
			# get new parent track
			drop_track = self.get_valid_drop(item.get_bounds().x1, item.get_bounds().y1)
			
			# update clip's settings to reflect the new x, y position (and update the parent track)
			self.update(item.get_bounds().x1, item.get_bounds().y1, drop_track)
			drop_track.reorder_clips()  
	
			# deterime the direction of the drag
			if isSnap:
				distance_from_clip = self.get_snap_difference(self, item)
			else:
				distance_from_clip = 0
				
			# move the clip object on the timeline to correct position (snapping to the y and x of the track)		
			item.translate (distance_from_clip, drop_track.y_top - item.get_bounds().y1 + 2)
			
			# update clip's settings again (since the snapping could have moved them a bit)
			self.update(item.get_bounds().x1, item.get_bounds().y1, drop_track)
			
			# Protect the visible button from track 0
			self.protect_visible_button()
	
			
		elif isResize:
			# remove clip from goocanvas
			parent = item.get_parent()
			child_num = parent.find_child (item)
			parent.remove_child (child_num)			
			
			# re-render the clip at it's new size
			self.RenderClip()
			
			
		# reset the cursor
		self.parent.parent.project.form.MyCanvas.window.set_cursor(None)

		
		
	def is_valid_drop(self, x1, y1):
		""" Check to see if the current position is a valid drop position """

		# loop through each track
		for track in self.parent.parent.tracks:

			# get the top y and bottom y of each track
			y_top = track.y_top
			y_bottom = track.y_bottom
			
			# get the middle of the clip
			half_height_of_clip = 26
			middle_position = half_height_of_clip + y1
			
			# determine if middle of clip is contained inside this track
			if middle_position > y_top and middle_position < y_bottom:
				return True

		# return false if no valid track found
		return False
	
	
	def get_valid_drop(self, x1, y1):
		""" A clip must be dropped on a track.  This method returns the track 
		object that is under the clip's current position """

		# loop through each track
		for track in self.parent.parent.tracks:
			# get the top y and bottom y of each track
			y_top = track.y_top
			y_bottom = track.y_bottom
			
			# get the middle of the clip
			half_height_of_clip = 26
			middle_position = half_height_of_clip + y1
			
			# determine if middle of clip is contained inside this track
			if middle_position > y_top and middle_position < y_bottom:
				return track
		
		# return false if no valid track found
		return None
		
		
	def on_visible_click (self, item, target, event):
		# get a reference to the 2 main canvas objects & theme
		theme = self.parent.parent.project.theme
		
		# get the parent left group
		parent_group = item.get_parent()

		if self.play_video == True:
			# Load Hover Over
			imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/not_visible_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
			item.set_properties(pixbuf = imgTrack_Visible.get_pixbuf())

			# update play video variable
			self.play_video = False
			
		else: 
			# Load normal image
			imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/visible_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
			item.set_properties(pixbuf = imgTrack_Visible.get_pixbuf())

			# update play video variable
			self.play_video = True

		return False
	
	
	def on_audio_click (self, item, target, event):
		# get a reference to the 2 main canvas objects & theme
		theme = self.parent.parent.project.theme
		
		# get the parent left group
		parent_group = item.get_parent()

		if self.play_audio == True:
			# Load Hover Over
			imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/speaker_mute_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
			item.set_properties(pixbuf = imgTrack_Visible.get_pixbuf())
			
			# update play video variable
			self.play_audio = False
			
		else: 
			# Load normal image
			imgTrack_Visible = gtk.image_new_from_file("%s/themes/%s/speaker_transparent.png" % (self.parent.parent.project.form.openshot_path, theme))
			item.set_properties(pixbuf = imgTrack_Visible.get_pixbuf())
			
			# update play video variable
			self.play_audio = True

		return False
		
		
		
		
		
		
		
		
		
		
		