#	OpenShot Video Editor is a program that creates, modifies, and edits video files.
#   Copyright (C) 2009  Jonathan Thomas
#
#	This file is part of OpenShot Video Editor (http://launchpad.net/openshot/).
#
#	OpenShot Video Editor is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	OpenShot Video Editor is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with OpenShot Video Editor.  If not, see <http://www.gnu.org/licenses/>.

import track
import timeline
import marker
import goocanvas
import gtk
import xml.dom.minidom as xml

########################################################################
class sequence:
	"""A sequence contains tracks and clips that make up a scene (aka sequence).  Currently, Openshot
	only contains a single sequence, but soon it will have the ability to create many sequences."""

	#----------------------------------------------------------------------
	def __init__(self, seq_name, project):
		"""Constructor"""
		
		# init variables for sequence
		self.name = seq_name
		self.length = 600		   # length in seconds of sequence.  This controls how wide to render the tracks.
		self.project = project	  # reference to current project
		self.scale = 15.0		   # this represents how many seconds per tick mark
		self.tick_pixels = 100	  # number of pixels between each tick mark
		self.play_head_position = 0 # position of the play head in seconds
		
		# init the tracks on the sequence
		self.tracks = []
		self.tracks.append(track.track("Track 2", self))
		self.tracks.append(track.track("Track 1", self))
		
		# init markers
		self.markers = []
		
		# reference to play_head goocanvas group
		self.play_head = None
		self.ruler_time = None
		
		
	def AddMarker(self, marker_name, position_on_track):
		
		# loop through existing markers (check for duplicates)
		for exiting_marker in self.markers:
			if exiting_marker.position_on_track == position_on_track:
				return
		
		# Create marker object
		m = marker.marker(marker_name, position_on_track, self)
		
		# Add to Marker Collection
		self.markers.append(m)
		
		# re-order marker collection (to keep them in order of their positions)
		self.reorder_markers()
		
		return m

		

	def get_marker(self, direction, current_position):
		

		# check for markers
		if len(self.markers) == 0:
			return None

		if direction == "left":
			
			# loop through all markers (check for duplicates)
			possible_match = None
			for m in self.markers:

				if m.position_on_track < current_position - 0.20:
					possible_match = m

			# return the match
			return possible_match
			
		else:
			# right
			
			# loop through all markers (check for duplicates)
			for m in self.markers:

				if m.position_on_track > current_position + 0.20:
					# return the match
					return m
				
				
		# no markers found
		return None
		
		
		
	def AddTrack(self, track_name, position="above", existing_track=None):
		
		# Add another track
		if existing_track:
			index_existing_track = self.tracks.index(existing_track)
			# insert next to an existing track
			if position == "above":
				self.tracks.insert(index_existing_track, track.track(track_name, self))
			else:
				self.tracks.insert(index_existing_track + 1, track.track(track_name, self))
				
		else:
			# insert at top of tracks
			self.tracks.insert(0, track.track(track_name, self))

		
	def get_pixels_per_second(self):
		# calculate the number of pixels per second, based on the scale of the timeline
		return float(self.tick_pixels) / float(self.scale)	   
		
	
	def Render(self):

		# Clear the canvases
		self.project.form.MyCanvas_Left.set_root_item(goocanvas.Group())
		self.project.form.MyCanvas.set_root_item(goocanvas.Group())
		
		# Render Ruler
		self.RenderRuler()
		
		# Render Markers
		self.RenderMarkers()
		
		# loop through each track
		for MyTrack in self.tracks:
			
			# Render track			
			MyTrack.Render()
			
			
	def GenerateXML(self, dom, xmlParentNode):
		
		multitrack = dom.createElement("multitrack")
		xmlParentNode.appendChild(multitrack)
		
		# loop through each track, from the bottom up
		for MyTrack in reversed(self.tracks):
			
			# Render track			
			MyTrack.GenerateXML(dom, multitrack)

			
			
	#----------------------------------------------------------------------
	def RenderRuler(self):
		"""This adds a track to the canvas with 3 images: a left, middle, and right"""
		
		# get the pixels per second from the parent sequence
		pixels_per_second = self.get_pixels_per_second()
		
		# determine position of playhead
		time = timeline.timeline().get_friendly_time(self.play_head_position * 1000) 
		
		# set the top coordinate of this track to the bottom coordinate of the last track
		x = 10	  
		y_top = 10
		
		# get a reference to the 2 main canvas objects & theme
		theme = self.project.theme
		canvas_left = self.project.form.MyCanvas_Left
		canvas_right = self.project.form.MyCanvas
		
		# Add an item to the goocanvas
		root_left = canvas_left.get_root_item ()
		root_right = canvas_right.get_root_item ()

		# Load all 3 images
		imgTrack_Left = gtk.image_new_from_file("%s/themes/%s/ruler_left.png" % (self.project.form.openshot_path, theme))
		imgTrack_Middle = gtk.image_new_from_file("%s/themes/%s/ruler_middle.png" % (self.project.form.openshot_path, theme))
		imgTrack_Right = gtk.image_new_from_file("%s/themes/%s/ruler_right.png" % (self.project.form.openshot_path, theme))	   
		imgTrack_Tick = gtk.image_new_from_file("%s/themes/%s/ruler_tick.png" % (self.project.form.openshot_path, theme))	   

		# Get Height and Width of Images 
		imgTrack_Left_Height = imgTrack_Left.get_pixbuf().get_height()
		imgTrack_Left_Width = imgTrack_Left.get_pixbuf().get_width()
		imgTrack_Right_Width = imgTrack_Right.get_pixbuf().get_width()		

		# Get Size of Window (to determine how wide the middle image should be streched)
		Size_Of_Middle = int(pixels_per_second * self.length)
		
		# resize the canvas height and width
		canvas_right.set_bounds (0, 0, Size_Of_Middle + 100, 1000)
		self.project.form.hscrollbar2.set_range(0, Size_Of_Middle + 100)
		self.project.form.on_hscrollbar2_value_changed(self.project.form.hscrollbar2)

		# Resize Middle pixbuf
		middlePixBuf = imgTrack_Middle.get_pixbuf()
		pixbuf_list = self.split_images(middlePixBuf, imgTrack_Left_Height, Size_Of_Middle)
		
		# Create Group (for the track)
		GroupTrack = goocanvas.Group (parent = root_right)

		# Add Left Image to Group
		image1 = goocanvas.Image (parent = root_left,
								  pixbuf = imgTrack_Left.get_pixbuf(),
								  x = x,
								  y = y_top)

		# Add Middle Image to Group (this can be multiple image tiled together
		pixbuf_x = 0
		for pixbuf in pixbuf_list:
			# get width of this pixbuf
			pixbuf_width = pixbuf.get_width()
			
			image2 = goocanvas.Image (parent = GroupTrack,
									  pixbuf = pixbuf,
									  x = pixbuf_x,
									  y = y_top)  
			
			# increment the x
			pixbuf_x = pixbuf_x + pixbuf_width


		# Add Middle Image to Group
		image3 = goocanvas.Image (parent = GroupTrack,
								  pixbuf = imgTrack_Right.get_pixbuf(),
								  x = Size_Of_Middle - 1,
								  y = y_top)

		# Add Text to the Track
		self.ruler_time = goocanvas.Text (parent = root_left,
								text = "%02d:%02d:%02d:%03d" % (time[2], time[3], time[4], time[5]),
								font = "Sans 12",
								antialias = True,
								x = x + 27,
								y = y_top + 8)
			  
		# Resize tick marks for ruler
		tickPixBuf = imgTrack_Tick.get_pixbuf()
		big_tickPixBuf = tickPixBuf.scale_simple(1, 24, gtk.gdk.INTERP_NEAREST)
		medium_tickPixBuf = tickPixBuf.scale_simple(1, 16, gtk.gdk.INTERP_NEAREST)
		small_tickPixBuf = tickPixBuf.scale_simple(1, 10, gtk.gdk.INTERP_NEAREST)
		
		# loop through each tick mark
		number_of_ticks = int(self.length / self.scale)
		for tick in range(0, number_of_ticks):
			if tick > 0:
				
				# get the formatted time code
				milliseconds = (tick * self.scale) * 1000
				time = timeline.timeline().get_friendly_time(milliseconds)
				
				# Add Text to the Track
				text1 = goocanvas.Text (parent = GroupTrack,
										text = "<span face='Sans' foreground='#2A5799' size='7000'>%02d:%02d:%02d:%03d</span>" % (time[2], time[3], time[4], time[5]),
										
										antialias = True,
										use_markup = True,
										x = (tick * self.tick_pixels) - 31,
										y = y_top + 3)				
				
				# Add BIG TICK Image to Group
				image5 = goocanvas.Image (parent = GroupTrack,
										  pixbuf = big_tickPixBuf,
										  x = (tick * self.tick_pixels),
										  y = y_top + 15)
				
				# determine # of minor ticks.  The longer the timeline, the less
				# ticks can be added do to performance reasons.
				number_of_minor_ticks = 6
				if (number_of_ticks * 10) > 1500:
					number_of_minor_ticks = 2
					
				# Add Minor ticks
				big_tick_x = (tick * self.tick_pixels)
				minor_tick_increment = self.tick_pixels / number_of_minor_ticks
 
				# loop through minor ticks, and add them to the canvas
				for minor_tick in range(1, number_of_minor_ticks):

					# Resize minor tick pixbuf
					tickPixBuf = None
					minor_y_adjustment = 29
					if minor_tick == int(number_of_minor_ticks / 2):
						minor_y_adjustment = 23
						tickPixBuf = medium_tickPixBuf
					else:
						tickPixBuf = small_tickPixBuf
					
					# Add minor tick Image to Group
					image5 = goocanvas.Image (parent = GroupTrack,
											  pixbuf = tickPixBuf,
											  x = big_tick_x - int(minor_tick_increment * minor_tick),
											  y = y_top + minor_y_adjustment)
					
	
	#----------------------------------------------------------------------
	def RenderMarkers(self):
		
		# loop through all markers (check for duplicates)
		for m in self.markers:
			
			# render the marker
			m.Render()
			
				
				
	def reorder_markers(self):
		# get a list of all clips on this track
		self.markers.sort(self.compare_marker)

		
		
	def compare_marker(self, marker1, marker2):
		if marker1.position_on_track > marker2.position_on_track:
			return 1
		elif marker1.position_on_track == marker2.position_on_track:
			return 0
		else:
			return -1
				
				
	#----------------------------------------------------------------------
	def RenderPlayHead(self):
		"""This adds the playhead to the canvas, and the play head position line"""
		
		# get the pixels per second from the parent sequence
		pixels_per_second = self.get_pixels_per_second()
		
		# set the top coordinate of this track to the bottom coordinate of the last track
		x = self.play_head_position * pixels_per_second
		y_top = 9
		
		# get a reference to the 2 main canvas objects & theme
		theme = self.project.theme
		canvas_right = self.project.form.MyCanvas
		
		# Add an item to the goocanvas
		root_right = canvas_right.get_root_item ()

		# Load all 3 images
		imgTrack_PlayHead = gtk.image_new_from_file("%s/themes/%s/play_head.png" % (self.project.form.openshot_path, theme))
		imgTrack_Line = gtk.image_new_from_file("%s/themes/%s/position_line.png" % (self.project.form.openshot_path, theme))
		imgTrack_Ruler = gtk.image_new_from_file("%s/themes/%s/ruler_right.png" % (self.project.form.openshot_path, theme))
		imgTrack_Track = gtk.image_new_from_file("%s/themes/%s/Track_Middle.png" % (self.project.form.openshot_path, theme))

		# Get Height and Width of Images 
		imgTrack_PlayHead_Height = imgTrack_PlayHead.get_pixbuf().get_height()
		imgTrack_PlayHead_Width = imgTrack_PlayHead.get_pixbuf().get_width()
		imgTrack_Ruler_Height = imgTrack_Ruler.get_pixbuf().get_height()
		imgTrack_Track_Height = imgTrack_Track.get_pixbuf().get_height()

		# Get Size of Window (to determine how wide the middle image should be streched)
		Size_Of_Line = len(self.tracks) * imgTrack_Track_Height + len(self.tracks * 6) + imgTrack_Ruler_Height - 4

		# Resize Middle pixbuf
		linePixBuf = imgTrack_Line.get_pixbuf()
		linePixBuf = linePixBuf.scale_simple(1, Size_Of_Line, gtk.gdk.INTERP_NEAREST)
		
		# Create Group (for the track)
		GroupTrack = goocanvas.Group (parent = root_right)
		self.play_head = GroupTrack

		# Create the red line
		lineRect = goocanvas.Rect(parent = GroupTrack,
								  x = x,
								  y = y_top + 3,
								  width = 1,
								  height = Size_Of_Line,
								  line_width = .5,
								  stroke_color = 'Red'
								  )

		# Add Play Head Image to Group
		image1 = goocanvas.Image (parent = GroupTrack,
								  pixbuf = imgTrack_PlayHead.get_pixbuf(),
								  x = x + (imgTrack_PlayHead_Width / 2) * -1,
								  y = imgTrack_Ruler_Height - imgTrack_PlayHead_Height + y_top - 2)

		# Connect signals to clip to allow drag and drop
		GroupTrack.connect ("motion_notify_event", self.on_motion_notify_x)
		GroupTrack.connect ("button_press_event", self.on_button_press_x)
		GroupTrack.connect ("button_release_event", self.on_button_release_x)


		
	def on_motion_notify_x (self, item, target, event):
		"""this method allows the clip to be dragged and dropped on a track"""

		if (event.state & gtk.gdk.BUTTON1_MASK):
			new_x = event.x
			new_y = event.y
			
			# get the pixels per second from the parent sequence
			pixels_per_second = self.get_pixels_per_second()
			
			# get width of playhead
			playhead_width = item.get_bounds().x2 - item.get_bounds().x1
			min_x = (playhead_width / 2) * -1

			# don't allow the clip to slide past the beginning of the canvas
			total_x_diff = new_x - self.drag_x 
			total_y_diff = event.y - self.drag_y
			if (item.get_bounds().x1 + total_x_diff < min_x):
				total_x_diff = min_x - item.get_bounds().x1

			# move clip				
			item.translate (total_x_diff, 0)

			# print time of playhead
			milliseconds = ((item.get_bounds().x1 + (playhead_width / 2)) / pixels_per_second) * 1000
			time = timeline.timeline().get_friendly_time(milliseconds)
			self.ruler_time.set_property("text", "%02d:%02d:%02d:%03d" % (time[2], time[3], time[4], time[5]))

			# update playhead position
			self.play_head_position = (item.get_bounds().x1 + (playhead_width / 2)) / pixels_per_second
			
			# check for video object
			if self.project.form.MyVideo:
				# stop the video
				self.project.form.MyVideo.pause()
				
				# determine new frame number
				frame = self.project.fps() * self.play_head_position
				
				# seek to the new frame
				self.project.form.MyVideo.seek(int(frame))

		return True
	

	def move_play_head (self, new_time):
		"""this method allows the play head to be moved to a specific spot on the timeline.  It accepts
		a parameter for the # of seconds to move the playhead to."""
		
		# move play_head to the top layer
		self.play_head.raise_(None)

		# get the pixels per second from the parent sequence
		pixels_per_second = self.get_pixels_per_second()
		
		# get width of playhead
		playhead_width = self.play_head.get_bounds().x2 - self.play_head.get_bounds().x1
		min_x = (playhead_width / 2) * -1

		# don't allow the clip to slide past the beginning of the canvas
		total_x_diff = (new_time * pixels_per_second) - (self.play_head.get_bounds().x2 + min_x)

		# move clip				
		self.play_head.translate (total_x_diff, 0)
		
		# print time of playhead
		milliseconds = ((self.play_head.get_bounds().x1 + (playhead_width / 2)) / pixels_per_second) * 1000
		time = timeline.timeline().get_friendly_time(milliseconds)
		self.ruler_time.set_property("text", "%02d:%02d:%02d:%03d" % (time[2], time[3], time[4], time[5]))
		
		# update playhead position
		self.play_head_position = (self.play_head.get_bounds().x1 + (playhead_width / 2)) / pixels_per_second

	
	
	def on_button_press_x (self, item, target, event):
		""" This method initializes some variables needed for dragging and dropping a clip """
		# raise the group up to the top level
		item.raise_(None)

		if event.button == 1:

			# set the x and y where the cursor started dragging from
			self.drag_x = event.x
			self.drag_y = event.y

			# change the cursor for the drag n drop operation
			fleur = gtk.gdk.Cursor (gtk.gdk.FLEUR)
			canvas = item.get_canvas ()
			canvas.pointer_grab (item,
								 gtk.gdk.POINTER_MOTION_MASK | gtk.gdk.BUTTON_RELEASE_MASK,
								 fleur,
								 event.time)

		return True
	
	
	def on_button_release_x (self, item, target, event):
		""" This method drops a clip, and snaps the clip to the nearest valid track """
		
		# get reference to the canvas, and stop dragging the item
		canvas = item.get_canvas ()
		canvas.pointer_ungrab (item, event.time)
		
		
		
		
	def split_images(self, pixbuf, height, max_length):
		""" Because it's not possible to resize an image to an infinate size, we sometimes
		need to split an image into many smaller pieces.  This function takes an image, and
		returns a list of pixbufs to equal the max_length"""
		
		# how big of chunks do we use (pixels)
		chunk_size = 1000
		image_list = []
		
		# get a reference to the 2 main canvas objects & theme
		remaining_pixels = max_length
		
		while remaining_pixels:
			# declare new image
			new_pixbuf = None

			if chunk_size < remaining_pixels:
				# resize the pixbuf
				new_pixbuf = pixbuf.scale_simple(chunk_size, height, gtk.gdk.INTERP_NEAREST)
				
				# subtract the pixels
				remaining_pixels = remaining_pixels - chunk_size
				
			else:
				# resize the pixbuf
				new_pixbuf = pixbuf.scale_simple(remaining_pixels, height, gtk.gdk.INTERP_NEAREST)
		
				# subtract the pixels
				remaining_pixels = 0
				
			# add image to list
			image_list.append(new_pixbuf)
		
		# return list of pixbufs
		return image_list

		
		
		
		
		
		
		
		